/*-
 * Copyright (c) 2012 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "bigint.h"

int
main(int argc, char **argv)
{
	uint4096_t	stack[32];
	uint4096_t	tmp;
	uint64_t	bits;
	size_t		nbits;
	char		buf[BUFSIZ];
	int		stackc;
	int		base;
	int		from;
	int		len;
	int		ok;
	int		i;

	nbits = 4096;
	while ((i = getopt(argc, argv, "n:")) != -1) {
		switch(i) {
		case 'n':
			nbits = (size_t)strtoul(optarg, NULL, 0);
			break;
		default:
			break;
		}
	}
	stackc = 0;
	for (;;) {
		if (isatty(0)) {
			fprintf(stderr, "fpcalc: ");
		}
		if (fgets(buf, sizeof(buf), stdin) == NULL) {
			break;
		}
		switch(buf[0]) {
		case 'P':
		case '=':
			bigint_fprintf(stdout, &stack[stackc - 1], nbits);
			printf("\n");
			break;
		case 'p':
			if (stackc > 0) {
				stackc -= 1;
			}
			break;
		case '+':
		case '-':
		case '*':
		case '/':
		case '%':
		case 'g':
			if (stackc < 2) {
				fprintf(stderr, "STACK UNDERFLOW\n");
				continue;
			}
			ok = 1;
			switch(buf[0]) {
			case '+':
			case '-':
			case '*':
			case '/':
			case '%':
				ok = bigint_arith(&tmp, &stack[stackc - 1], buf[0], &stack[stackc - 2], nbits);
				break;
			case 'g':
				if (!bigint_gcd(&tmp, &stack[stackc - 1], &stack[stackc - 2], nbits)) {
					ok = 0;
				}
				break;
			default:
				break;
			}
			if (ok) {
				(void) memcpy(&stack[stackc-- - 2], &tmp, sizeof(tmp));
			}
			break;
		case 'l':
			if (bigint_uint64(&stack[stackc - 1], &bits)) {
				bigint_word_arith(&tmp, &stack[stackc - 2], 'l', bits, nbits);
				(void) memcpy(&stack[stackc-- - 2], &tmp, sizeof(tmp));
			}
			break;
		case 'r':
			if (bigint_uint64(&stack[stackc - 1], &bits)) {
				bigint_word_arith(&tmp, &stack[stackc - 2], 'r', bits, nbits);
				(void) memcpy(&stack[stackc-- - 2], &tmp, sizeof(tmp));
			}
			break;
		default:
			if (stackc == __arraycount(stack) - 1) {
				fprintf(stderr, "STACK OVERFLOW\n");
				continue;
			}
			from = 0;
			len = strlen(buf) - 1;
			base = 10;
			if (strncmp(buf, "0x", 2) == 0) {
				base = 16;
				from = 2;
				len -= 2;
			} else if (buf[strlen(buf) - 2] == 'h') {
				base = 16;
				len -= 1;
			}
			bigint_str2bigint(&stack[stackc++], &buf[from], len, base, nbits);
			break;
		}
	}
	exit(EXIT_SUCCESS);
}
